from netgen.csg import unit_cube
from ngsolve import *

mesh = Mesh(unit_cube.GenerateMesh(maxh=0.1))
Draw(mesh)

fes = HCurl(mesh, order=1, type1=True)
u,v = fes.TnT()

a = BilinearForm(fes, symmetric=True)
a += u*v*dx
a.Assemble()

f = LinearForm(fes)
f += CoefficientFunction( (x,y,1) )*v*dx
f.Assemble()

gfu = GridFunction(fes)

gfu.vec.data = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky")*f.vec
print("err (lo) =", Integrate((gfu - CoefficientFunction( (x,y,1) ))**2,mesh))

###########################################################
fes.SetOrder(TRIG,2)
fes.SetOrder(TET,3)
fes.Update()
u,v = fes.TnT()

a = BilinearForm(fes, symmetric=True)
a += u*v*dx
a.Assemble()

f = LinearForm(fes)
f += CoefficientFunction( (x,y,1) )*v*dx
f.Assemble()

gfu.Update()
gfu.vec.data = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky")*f.vec
print("err (lo + face + inner) =", Integrate((gfu - CoefficientFunction( (x,y,1) ))**2,mesh))

##########################################################
print(sum(fes.FreeDofs()))
bitarray = fes.FreeDofs()
for i in range(mesh.nedge):
    bitarray[i] = 0
print(sum(bitarray))

gfu.vec.data = a.mat.Inverse(bitarray, inverse="sparsecholesky")*f.vec
print("err (face + inner), bitarray =", Integrate((gfu - CoefficientFunction( (x,y,1) ))**2,mesh))


##########################################################
for i in range(mesh.nedge):
    fes.SetCouplingType(i, COUPLING_TYPE.UNUSED_DOF)

gfu.vec.data = a.mat.Inverse(fes.FreeDofs(), inverse="sparsecholesky")*f.vec
print("err (face + inner), UNUSED_DOF =", Integrate((gfu - CoefficientFunction( (x,y,1) ))**2,mesh))
